from sklearn.pipeline import Pipeline

from ... import nodes as global_nodes
from ...abstract import AbstractPipelineFactory
from ...models import CRSProjectionModel
from ...models import DataFramePivotFields


class PreprocessingFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline(
        source_lat_column: str,
        source_lon_column: str,
        source_datetime: str,
        source_vehicle_id: str,
        source_crs: str,
        target_crs: str,
    ) -> Pipeline:
        """Function creating instance of fragmentation module

        Args:
        source_lat_column (str): source latitude
        source_lon_column (str): source longitude
        source_datetime (str): datetime
        source_crs (str): source coordinate reference system
        target_crs (str): target coordinate reference system
        max_velocity_hard_limit (Union[float, int]): maximum velocity
            to consider as cluster candidate

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """
        pivot_fields = DataFramePivotFields(
            source_lat=source_lat_column,
            source_lon=source_lon_column,
            source_datetime=source_datetime,
            source_vehicle_id=source_vehicle_id,
        )
        projection_model = CRSProjectionModel(
            source_crs=source_crs,
            target_crs=target_crs,
        )
        return Pipeline(
            [
                (
                    "schema_validator",
                    global_nodes.PanderaValidator(pivot_fields=pivot_fields),
                ),
                (
                    "unique_vehicle_constrain",
                    global_nodes.UniqueVehicleConstrain(pivot_fields=pivot_fields),
                ),
                (
                    "crs_projector",
                    global_nodes.CRSTransformer(
                        crs_projection=projection_model,
                        pivot_fields=pivot_fields,
                    ),
                ),
                (
                    "unixtime_extractor",
                    global_nodes.UnixtimeExtractor(
                        source_column=pivot_fields.source_datetime,
                        target_column=pivot_fields.computed_unixtime,
                    ),
                ),
                (
                    "velocity_calculator",
                    global_nodes.VelocityCalculator(pivot_fields=pivot_fields),
                ),
            ],
        )



import logging
import numpy as np
import pandas as pd

from ...abstract import AbstractNode
from ...models import DataFramePivotFields
from ...models import DefaultValues


default_values = DefaultValues()
pivot_fields = DataFramePivotFields()
_VALID_FLAG_COL = pivot_fields.sjoin_valid_flag
_GPS_SUFFIX = default_values.sjoin_gps_suffix
_PLAN_SUFFIX = default_values.sjoin_plan_suffix


class SpatialJoinValidator(AbstractNode):
    """
    Module performing validation based on proximity between plan & gps locations
    NOTE: here is a corner case when no joins are found between GPS & Plan
    """

    DEFAULT_EMPTY_DATAFRAME = pd.DataFrame(
        {
            pivot_fields.sjoin_temporal_dist: [],
            pivot_fields.sjoin_spatial_dist: [],
            pivot_fields.sjoin_overall_dist: [],
        },
    )

    def __init__(
        self,
        max_days_distance: int,
        ensure_vehicle_overlap: bool = True,
    ):
        self.pivot_fields = DataFramePivotFields()
        self.ensure_vehicle_overlap = ensure_vehicle_overlap
        self.max_days_distance = max_days_distance

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def __init_validation_flag(self, X: pd.DataFrame) -> pd.DataFrame:
        X[_VALID_FLAG_COL] = True
        return X

    def __interpolate_colname(self, column: str, suffix: str) -> str:
        return f"{column}_{suffix}"

    @property
    def gps_date(self):
        date_col = self.pivot_fields.projected_date
        return self.__interpolate_colname(date_col, suffix=_GPS_SUFFIX)

    @property
    def plan_date(self):
        date_col = self.pivot_fields.projected_date
        return self.__interpolate_colname(date_col, suffix=_PLAN_SUFFIX)

    @property
    def gps_vehicle_id(self):
        vehicle_id = self.pivot_fields.source_vehicle_id
        return self.__interpolate_colname(vehicle_id, suffix=_GPS_SUFFIX)

    @property
    def plan_vehicle_id(self):
        vehicle_id = self.pivot_fields.source_vehicle_id
        return self.__interpolate_colname(vehicle_id, suffix=_PLAN_SUFFIX)

    def __calculate_days_distance(self, X: pd.DataFrame, gps_date: str, plan_date: str):
        delta_days = (X[gps_date] - X[plan_date]).dt.days
        return np.abs(delta_days)

    def __refresh_validation_flag(self, X: pd.DataFrame, validation_mask: pd.Series):
        X[_VALID_FLAG_COL] = X[_VALID_FLAG_COL] & validation_mask
        return X

    def __validate_temporal_tolerance(self, X: pd.DataFrame) -> pd.DataFrame:
        delta_days = self.__calculate_days_distance(
            X,
            gps_date=self.gps_date,
            plan_date=self.plan_date,
        )
        X[self.pivot_fields.sjoin_temporal_dist] = delta_days
        is_date_validated = delta_days <= self.max_days_distance
        X = self.__refresh_validation_flag(X, is_date_validated)
        return X

    def __validate_vehicle_overlap(self, X: pd.DataFrame) -> pd.DataFrame:
        are_vehicles_match = X[self.gps_vehicle_id] == X[self.plan_vehicle_id]
        X = self.__refresh_validation_flag(X, are_vehicles_match)
        return X

    def __apply_validations(self, X: pd.DataFrame) -> pd.DataFrame:
        X = self.__init_validation_flag(X)
        X = self.__validate_temporal_tolerance(X)
        X = self.__validate_vehicle_overlap(X)
        return X

    def __drop_joins_failed_validation(self, X: pd.DataFrame) -> pd.DataFrame:
        X = X.loc[X[self.pivot_fields.sjoin_valid_flag], :]
        X = X.reset_index(drop=True)
        return X

    def __get_temp_dist_weight(self, X: pd.DataFrame):
        return X[self.pivot_fields.sjoin_spatial_dist].max()

    def __get_spatial_dist_weight(self):
        return 1.0

    def __compute_weighted_overall_distance(self, X: pd.DataFrame) -> pd.DataFrame:
        _fields = self.pivot_fields
        X[_fields.sjoin_overall_dist] = (X[_fields.sjoin_temporal_dist] * self.__get_temp_dist_weight(X)) + (
            X[_fields.sjoin_spatial_dist] * self.__get_spatial_dist_weight()
        )
        return X

    def __sort_by_overall_distance(self, X: pd.DataFrame) -> pd.DataFrame:
        return X.sort_values(by=[self.pivot_fields.sjoin_overall_dist]).reset_index(drop=True)

    def __ensure_cardinality(self, X: pd.DataFrame):
        """
        Many jobs can be assigned to only one cluster
        One cluster can't be assinged to many job
        """
        return X.drop_duplicates(subset=[self.pivot_fields.plans_pk])

    def __is_has_any_cluster_plan_join(self, X: pd.DataFrame):
        return X.shape[0] > 0

    def __expand_empty_input_with_keys(self, X: pd.DataFrame):
        return pd.concat([X, self.DEFAULT_EMPTY_DATAFRAME], axis=1)

    def transform(self, X: pd.DataFrame):
        if self.__is_has_any_cluster_plan_join(X):
            X = self.__apply_validations(X)
            X = self.__drop_joins_failed_validation(X)
            X = self.__compute_weighted_overall_distance(X)
            X = self.__sort_by_overall_distance(X)
            X = self.__ensure_cardinality(X)
        else:
            logging.error("Provided session does't have any cluster-plan join")
            X = self.__expand_empty_input_with_keys(X)

        return X



from typing import List, Literal
from sklearn.pipeline import Pipeline

from ... import nodes as global_nodes
from ...abstract import AbstractPipelineFactory
from ...models import CRSProjectionModel
from ...models import DataFramePivotFields


_DEFAULTS = DataFramePivotFields()
_PK_MAPPINGS = {
    "gps": _DEFAULTS.gps_pk,
    "plan": _DEFAULTS.plans_pk,
}


class PreprocessingFactory(AbstractPipelineFactory):
    @staticmethod
    def validate_primary_key_literal(generate_primary_key_for):
        if generate_primary_key_for not in _PK_MAPPINGS.keys():
            raise ValueError(
                "Incorrect generate_primary_key_for which " f"must be one from {_PK_MAPPINGS.keys()}",
            )

    @staticmethod
    def choose_pk_field(generate_primary_key_for):
        return _PK_MAPPINGS[generate_primary_key_for]

    # flake8: noqa: CFQ002
    @staticmethod
    def factory_pipeline(
        source_lat_column: str,
        source_lon_column: str,
        source_datetime: str,
        source_vehicle_id: str,
        source_crs: str,
        target_crs: str,
        generate_primary_key_for: str,
        source_composite_keys: List[str],
    ) -> Pipeline:
        """
        Function creating instance of preprocessing
        instance pipeline of linker module

        Args:
        source_lat_column: source latitude
        source_lon_column: source longitude
        source_datetime: datetime
        source_vehicle_id: source vehicle identifier key
        source_crs: source coordinate reference system
        target_crs: target coordinate reference system
        generate_primary_key_for: determines for which datasource generate primary key
        source_composite_keys: source composite keys to generate primary key

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """

        PreprocessingFactory.validate_primary_key_literal(generate_primary_key_for)

        pivot_fields = DataFramePivotFields(
            source_lat=source_lat_column,
            source_lon=source_lon_column,
            source_datetime=source_datetime,
            source_vehicle_id=source_vehicle_id,
        )
        default_pivot_fields = DataFramePivotFields()

        projection_model = CRSProjectionModel(
            source_crs=source_crs,
            target_crs=target_crs,
        )
        pk_field = PreprocessingFactory.choose_pk_field(
            generate_primary_key_for=generate_primary_key_for,
        )
        return Pipeline(
            [
                (
                    "schema_validator",
                    global_nodes.PanderaValidator(pivot_fields=pivot_fields),
                ),
                (
                    "primary_key_generator",
                    global_nodes.PrimaryKeyGenerator(
                        target_column=pk_field,
                        source_columns=source_composite_keys,
                    ),
                ),
                (
                    "default_fields_projector",
                    global_nodes.DefaultSchemaProjector(
                        original_pivot_fields=pivot_fields,
                    ),
                ),
                (
                    "date_extraction",
                    global_nodes.DateExtractor(
                        pivot_fields=default_pivot_fields,
                    ),
                ),
                (
                    "crs_projector",
                    global_nodes.CRSTransformer(
                        crs_projection=projection_model,
                        pivot_fields=default_pivot_fields,
                    ),
                ),
                (
                    "dataframe_2_geodataframe",
                    global_nodes.ConverterDataFrame2GeoDataFrame(
                        pivot_fields=default_pivot_fields,
                        crs_projection=projection_model,
                    ),
                ),
            ],
        )



import numpy as np
import pandas as pd

from ..abstract import AbstractNode
from ..models import DataFramePivotFields


class VelocityCalculator(AbstractNode):
    """
    Calculates velocity column from coordinates & time
    NOTE: given module is constrained over 1 velocity calculator = 1 vehicle
    """

    def __init__(
        self,
        pivot_fields: DataFramePivotFields,
    ):
        self.pivot_fields = pivot_fields

    def fit(self, X, y=None):
        return self

    def __get_l2_between_adjacent_points(self, X: pd.DataFrame) -> pd.DataFrame:
        lat_lon_columns = [self.pivot_fields.projected_lat, self.pivot_fields.projected_lon]
        lag_0_points = X.loc[:, lat_lon_columns]
        lag_1_points = X.loc[:, lat_lon_columns].shift(1)
        return lag_0_points - lag_1_points

    def __ensure_lack_time_duplicates(self, dt: pd.Series):
        if (dt == 0).any():
            raise ZeroDivisionError("Duplicated timestamps are met during computing velocity")

    def __get_unixtime_delta(self, X: pd.DataFrame) -> pd.DataFrame:
        unixtime_column = self.pivot_fields.computed_unixtime
        dt = X[unixtime_column] - X[unixtime_column].shift(1)
        dt = pd.Series(np.abs(dt))
        self.__ensure_lack_time_duplicates(dt)
        return dt

    def __sort_gps(self, X: pd.DataFrame) -> pd.DataFrame:
        sort_columns = [
            self.pivot_fields.source_vehicle_id,
            self.pivot_fields.computed_unixtime,
        ]
        X = X.sort_values(by=sort_columns)
        X.reset_index(drop=True, inplace=True)
        return X

    def __calculate_l2(self, dS: pd.DataFrame) -> pd.DataFrame:
        return np.sqrt((dS**2).T.sum())

    def __assign_velocity(
        self,
        X: pd.DataFrame,
        l2: pd.Series,
        dt: pd.Series,
    ) -> pd.DataFrame():
        X[self.pivot_fields.computed_velocity] = l2 / dt
        return X

    def __fillna_first(self, X: pd.DataFrame) -> pd.DataFrame:
        velocity_column = self.pivot_fields.computed_velocity
        X[velocity_column] = X[velocity_column].fillna(method="bfill")
        return X

    def __partition_over_vehicle_id(self, X: pd.DataFrame) -> pd.DataFrame:

        partitions = []
        vehicle_id_col = self.pivot_fields.source_vehicle_id

        for vehicle in X[vehicle_id_col].unique():
            _is_required_vehicle = X[vehicle_id_col] == vehicle
            _partition = X[_is_required_vehicle].reset_index(drop=True)
            partitions.append(_partition.copy())

        return partitions

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:

        X = self.__sort_gps(X)
        X = self.__partition_over_vehicle_id(X)

        dt = list(map(self.__get_unixtime_delta, X))
        dS = list(map(self.__get_l2_between_adjacent_points, X))

        l2 = list(map(self.__calculate_l2, dS))
        X = list(map(self.__assign_velocity, X, l2, dt))
        X = list(map(self.__fillna_first, X))

        X = pd.concat(X).reset_index(drop=True)

        return X



from typing import List
import pandas as pd

from ..abstract import AbstractNode
from ..models import DataFramePivotFields


class UniqueVehicleConstrain(AbstractNode):
    """
    Module validating constrain that only one unique
    vehicle must be listed in dataframe
    """

    def __init__(
        self,
        pivot_fields: DataFramePivotFields,
    ):
        self.pivot_fields = pivot_fields

    def fit(self, X, y=None):
        return self

    @property
    def vehicle_id(self):
        return self.pivot_fields.source_vehicle_id

    def __is_more_than_one_vehicle(self, X: pd.DataFrame) -> bool:
        return X[self.vehicle_id].nunique() > 1

    def __get_vehicles(self, X: pd.DataFrame) -> List[str]:
        return list(X[self.vehicle_id].unique())

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        if self.__is_more_than_one_vehicle(X):
            vehicles = self.__get_vehicles(X)
            message = f"More than one vehicle has been met in input data: {vehicles}"
            raise ValueError(message)
        return X



from .coordinates_velocity_calculator import VelocityCalculator
from .crs_projection import CRSTransformer
from .date_extractor import DateExtractor
from .df_to_gdf import ConverterDataFrame2GeoDataFrame
from .formatting import SelectColumns
from .primary_key_generator import PrimaryKeyGenerator
from .project_default_fields import DefaultSchemaProjector
from .schema_validator import PanderaValidator
from .timing import UnixtimeExtractor
from .unique_vehicle_constrains import UniqueVehicleConstrain


__all__ = [
    "SelectColumns",
    "CRSTransformer",
    "DateExtractor",
    "DefaultSchemaProjector",
    "UnixtimeExtractor",
    "PanderaValidator",
    "VelocityCalculator",
    "PrimaryKeyGenerator",
    "UniqueVehicleConstrain",
]



import pandas as pd

from ..abstract import AbstractNode
from ..models import DataFramePivotFields


class DateExtractor(AbstractNode):
    def __init__(
        self,
        pivot_fields: DataFramePivotFields,
    ):
        self.pivot_fields = pivot_fields

    def fit(self, X: pd.DataFrame, y=None):
        return self

    @property
    def datetime_column(self) -> str:
        return self.pivot_fields.source_datetime

    @property
    def date_column(self) -> str:
        return self.pivot_fields.projected_date

    def transform(self, X: pd.DataFrame):
        _datetimes = X[self.datetime_column]
        dates = pd.to_datetime(_datetimes).dt.date
        X[self.date_column] = dates
        return X



from typing import Literal
from pydantic import BaseModel


class Metrics(BaseModel):
    recall: float
    precision: float
    beta: float
    fbeta_score: float
    status: Literal["success", "fail"] = "success"
    status_details: str = ""

    @classmethod
    def factory_failure_metrics(cls, beta: str, status_details: str):
        return cls(
            recall=-1,
            precision=-1,
            beta=beta,
            fbeta_score=-1,
            status="fail",
            status_details=status_details,
        )



from typing import Dict, List, Union

from sklearn.pipeline import Pipeline


from ... import nodes as local_nodes
from .... import nodes as global_nodes
from ....abstract import AbstractPipelineFactory
from ....models import DataFramePivotFields


class FDBSCANFactory(AbstractPipelineFactory):

    # flake8: noqa: CFQ002
    @staticmethod
    def factory_pipeline(
        source_vehicle_id_column: str,
        eps: float = 0.5,
        min_samples: float = 5,
        metric: str = "euclidean",
        metric_params=None,
        algorithm="auto",
        leaf_size=30,
        p=None,
        n_jobs=-1,
    ) -> Pipeline:
        """
        Function builds scikit-learn clustering pipeline
        with fragmented density-based spatial clustering model.

        Built-in constraints:
        - Only one vehicle per pipeline run

        Args:
        eps: max distance to link a points to cluster
        min_samples: min amount of samples to form a cluster
        metric: metrics used for estimate
        metric_params: who knows
        algorithm: algorithm for model building
        leaf_size (int, optional): _description_. Defaults to 30.
        p (_type_, optional): _description_. Defaults to None.
        n_jobs (int, optional): _description_. Defaults to -1.

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn clustering pipeline
        """
        pivot_fields = DataFramePivotFields(source_vehicle_id=source_vehicle_id_column)
        return Pipeline(
            [
                (
                    "unique_vehicle_constrain",
                    global_nodes.UniqueVehicleConstrain(pivot_fields=pivot_fields),
                ),
                (
                    "select_columns",
                    global_nodes.SelectColumns(
                        columns=[
                            pivot_fields.projected_lat,
                            pivot_fields.projected_lon,
                            pivot_fields.computed_unixtime,
                            pivot_fields.fragmentation_output,
                        ],
                    ),
                ),
                (
                    "clustering_fdbscan",
                    local_nodes.FDBSCAN(
                        clustering_candidate_col=pivot_fields.fragmentation_output,
                        eps=eps,
                        min_samples=min_samples,
                        metric=metric,
                        metric_params=metric_params,
                        algorithm=algorithm,
                        leaf_size=leaf_size,
                        p=leaf_size,
                        n_jobs=n_jobs,
                    ),
                ),
            ],
        )



from sklearn.pipeline import Pipeline

from ... import nodes
from ....abstract import AbstractPipelineFactory
from ....models import DataFramePivotFields


pivot_fields = DataFramePivotFields()


_COMPUTED_VELOCITY_COLUMN = pivot_fields.computed_velocity


class VelocityFragmentationFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline(
        max_velocity_hard_limit: int,
    ) -> Pipeline:
        """
        Args:
        max_velocity_hard_limit (int): maximum speed limit to be clustering candidate

        Returns:
        Pipeline: Scikit learn pipeline
        """
        return Pipeline(
            [
                (
                    "velocity_fragmentation",
                    nodes.VelocityFragmentator(
                        source_velocity_column=_COMPUTED_VELOCITY_COLUMN,
                        max_velocity_hard_limit=max_velocity_hard_limit,
                    ),
                ),
            ],
        )



from sklearn.pipeline import Pipeline

from .. import nodes as local_nodes
from ... import nodes as global_nodes
from ...abstract import AbstractPipelineFactory
from ...models import CRSProjectionModel
from ...models import DataFramePivotFields


class ClusterAggregationFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline(
        source_lat_column: str,
        source_lon_column: str,
        source_datetime: str,
        source_vehicle_id: str,
        source_crs: str,
        target_crs: str,
    ) -> Pipeline:
        """
        Function creating instance of clustering
        instance pipeline of linker module

        Args:
        source_lat_column (str): source latitude
        source_lon_column (str): source longitude
        source_datetime (str): datetime
        source_vehicle_id (str): source vehicle identifier key
        source_crs (str): source coordinate reference system
        target_crs (str): target coordinate reference system

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """
        pivot_fields = DataFramePivotFields(
            source_lat=source_lat_column,
            source_lon=source_lon_column,
            source_datetime=source_datetime,
            source_vehicle_id=source_vehicle_id,
        )
        default_pivot_fields = DataFramePivotFields()
        projection_model = CRSProjectionModel(
            source_crs=source_crs,
            target_crs=target_crs,
        )
        return Pipeline(
            [
                (
                    "default_fields_projector",
                    global_nodes.DefaultSchemaProjector(
                        original_pivot_fields=pivot_fields,
                    ),
                ),
                (
                    "cluster_aggregator",
                    local_nodes.ClusterAggregator(default_pivot_fields),
                ),
                (
                    "dataframe_2_geodataframe",
                    global_nodes.ConverterDataFrame2GeoDataFrame(
                        pivot_fields=default_pivot_fields,
                        crs_projection=projection_model,
                    ),
                ),
                (
                    "primary_key_generator",
                    global_nodes.PrimaryKeyGenerator(
                        target_column=pivot_fields.clusters_pk,
                        source_columns=[
                            default_pivot_fields.source_vehicle_id,
                            default_pivot_fields.clustering_output,
                        ],
                    ),
                ),
            ],
        )



from typing import Dict, List, Union

from sklearn.pipeline import Pipeline


from ... import nodes as local_nodes
from .... import nodes as global_nodes
from ....abstract import AbstractPipelineFactory
from ....models import DataFramePivotFields


class STCMFactory(AbstractPipelineFactory):

    # flake8: noqa: CFQ002
    @staticmethod
    def factory_pipeline(
        source_vehicle_id_column: str,
        eps: float = 20,
        min_duration_sec: Union[int, float] = 60,
    ) -> Pipeline:
        """
        Function builds scikit-learn clustering pipeline
        with Spatio-temporal clustering model.

        Built-in constraints:
        - Only one vehicle per pipeline run

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn clustering pipeline
        """
        pivot_fields = DataFramePivotFields(source_vehicle_id=source_vehicle_id_column)
        return Pipeline(
            [
                (
                    "unique_vehicle_constrain",
                    global_nodes.UniqueVehicleConstrain(pivot_fields=pivot_fields),
                ),
                (
                    "select_columns",
                    global_nodes.SelectColumns(
                        columns=[
                            pivot_fields.projected_lat,
                            pivot_fields.projected_lon,
                            pivot_fields.computed_unixtime,
                            pivot_fields.fragmentation_output,
                        ],
                    ),
                ),
                (
                    "clustering_stcm",
                    local_nodes.STCM(
                        eps=eps,
                        min_duration_sec=min_duration_sec,
                    ),
                ),
            ],
        )



from ...abstract import AbstractNode
from ...models import DataContainer
from ...models import DataFramePivotFields


_PIVOT_FIELDS = DataFramePivotFields()
_PRIMARY_KEY = _PIVOT_FIELDS.clusters_pk


class Precision(AbstractNode):
    def fit(self, X: DataContainer, y=None):
        return self

    def transform(self, X: DataContainer):
        n_cluster_pkeys = X.clusters[_PRIMARY_KEY].nunique()
        n_join_pkeys = X.clusters_plan_join[_PRIMARY_KEY].nunique()
        return n_join_pkeys / n_cluster_pkeys



from typing import Any, Dict, List
import pandas as pd

from ...abstract import AbstractNode
from ...models import DataFramePivotFields
from ...models import DefaultValues


_GPS_NOISE_CLUSTER_ID = DefaultValues().noise_gps_cluster_id


class ClusterAggregator(AbstractNode):
    def __init__(
        self,
        pivot_fields: DataFramePivotFields,
        drop_noise_cluster: bool = True,
    ):
        self.pivot_fields = pivot_fields
        self.drop_noise_cluster = drop_noise_cluster

    def fit(self, X: pd.DataFrame, y=None):
        return self

    @property
    def group_columns(self) -> List[str]:
        return [
            self.pivot_fields.source_vehicle_id,
            self.pivot_fields.projected_date,
            self.pivot_fields.clustering_output,
        ]

    @property
    def agg_columns(self) -> Dict[str, Any]:
        return {
            self.pivot_fields.source_lon: (self.pivot_fields.source_lon, "mean"),
            self.pivot_fields.source_lat: (self.pivot_fields.source_lat, "mean"),
            self.pivot_fields.projected_lon: (self.pivot_fields.projected_lon, "mean"),
            self.pivot_fields.projected_lat: (self.pivot_fields.projected_lat, "mean"),
        }

    def __drop_noise_cluster(self, X: pd.DataFrame) -> pd.DataFrame:

        if self.drop_noise_cluster:
            cluster_column = self.pivot_fields.clustering_output
            is_not_noise_cluster = X[cluster_column] != _GPS_NOISE_CLUSTER_ID
            X = X[is_not_noise_cluster].reset_index(drop=True)

        return X

    def transform(self, X: pd.DataFrame):
        _agg = X.groupby(self.group_columns).agg(**self.agg_columns).reset_index()
        return self.__drop_noise_cluster(_agg)



from sklearn.pipeline import Pipeline

from .. import nodes as local_nodes
from ... import nodes as global_nodes
from ...abstract import AbstractPipelineFactory
from ...models import DataFramePivotFields
from ...models import DefaultValues


class JoinValidatorFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline(
        max_days_distance: int = 1,
        ensure_vehicle_overlap: bool = True,
    ) -> Pipeline:
        """
        Function creating instance of join
        validation pipeline of linker module

        Args:
        max_days_distance: Maximum temporal distance between
            plan and cluster join is allowed
        ensure_vehicle_overlap: if vehicle must be overlapped

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """
        default_fields = DataFramePivotFields()
        _default_values = DefaultValues()
        _gps_suffix = _default_values.sjoin_gps_suffix
        _plan_suffix = _default_values.sjoin_plan_suffix
        pivot_fields = DataFramePivotFields(
            source_lat=f"{default_fields.source_lat}_{_gps_suffix}",
            source_lon=f"{default_fields.source_lon}_{_gps_suffix}",
            projected_lat=f"{default_fields.source_lat}_{_plan_suffix}",
            projected_lon=f"{default_fields.source_lon}_{_plan_suffix}",
            source_datetime=f"{default_fields.projected_date}_{_gps_suffix}",
            source_vehicle_id=f"{default_fields.source_vehicle_id}_{_gps_suffix}",
        )
        return Pipeline(
            [
                (
                    "validator",
                    local_nodes.SpatialJoinValidator(
                        max_days_distance=max_days_distance,
                        ensure_vehicle_overlap=ensure_vehicle_overlap,
                    ),
                ),
                (
                    "schema_validator",
                    global_nodes.PanderaValidator(pivot_fields=pivot_fields),
                ),
                (
                    "columns_selector",
                    global_nodes.SelectColumns(
                        columns=[
                            pivot_fields.clusters_pk,
                            pivot_fields.plans_pk,
                            pivot_fields.sjoin_temporal_dist,
                            pivot_fields.sjoin_spatial_dist,
                            pivot_fields.sjoin_overall_dist,
                        ],
                    ),
                ),
            ],
        )



import pandas as pd

from . import nodes
from .models import Metrics
from ..abstract import AbstractNode
from ..models import DataContainer
from ..models import DataFramePivotFields


pivot_fields = DataFramePivotFields()


class ActivityMetricsSession(AbstractNode):
    """
    Class organizing sequence of computations of metrics of clustering data
    """

    DEFAULT_VALUE = -1
    NO_JOINS_STATUS_DETAILS = "No joins between clusters & plan"

    def __init__(
        self,
        beta: float = 1.0,
    ):
        """
        beta: beta argument parameter of fbeta score formula
        """
        self.beta = beta

    def fit(self, X, y=None):
        return self

    def __init_metrics_mapping(self):
        self.metrics_mappings = {
            "precision": nodes.Precision(),
            "recall": nodes.Recall(),
            "beta": self.beta,
            "fbeta_score": nodes.Fbeta(beta=self.beta),
        }

    def __compute_metric(self, X: pd.DataFrame, metric_key: str):
        self.metrics_mappings[metric_key] = self.metrics_mappings[metric_key].transform(X)

    def __is_has_any_cluster_plan_join(self, X: DataContainer):
        return X.clusters_plan_join.shape[0] > 0

    def transform(self, X: DataContainer) -> pd.DataFrame:
        if self.__is_has_any_cluster_plan_join(X):
            X.validated_coverage_stats()
            self.__init_metrics_mapping()
            self.__compute_metric(X, "precision")
            self.__compute_metric(X, "recall")
            self.__compute_metric(X, "fbeta_score")
            metrics = Metrics(**self.metrics_mappings)
        else:
            metrics = Metrics.factory_failure_metrics(
                beta=self.beta,
                status_details=self.NO_JOINS_STATUS_DETAILS,
            )

        return metrics



from typing import Literal, Union
from sklearn.pipeline import Pipeline

from .. import nodes as local_nodes
from ...abstract import AbstractPipelineFactory


class SpatialJoinerFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline(
        how: Literal["left", "right", "inner", "outer"],
        max_distance: Union[float, int],
    ) -> Pipeline:
        """
        Function creating instance of spatial join
        instance pipeline of linker module

        Args:
        how: join method
        max_distance: maximum allowed distance between points
            of left & right dataframe to make a spatial join

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """
        return Pipeline(
            [
                (
                    "cluster_aggregator",
                    local_nodes.SpatialJoiner(
                        how=how,
                        max_distance=max_distance,
                    ),
                ),
            ],
        )



from typing import Dict
import pandas as pd
from sklearn.pipeline import Pipeline

from .models import SpatialJoinArguments
from ..abstract import AbstractNode
from ..models import DataContainer
from ..models import DataFramePivotFields


pivot_fields = DataFramePivotFields()


class ActivityLinkageSession(AbstractNode):
    """
    Class organizing sequence of computations of linkage
    source gps clusters & route plan details.
    NOTE: here is a corner case when no joins are found between GPS & Plan
    """

    GPS = "gps"
    PLAN = "plan"

    # flake8: noqa: CFQ002
    def __init__(
        self,
        gps_preprocessor: Pipeline,
        plan_preprocessor: Pipeline,
        cluster_aggregator: Pipeline,
        spatial_joiner: Pipeline,
        spatial_validator: Pipeline,
        coverage_stats_extractor: Pipeline,
    ):
        """
        gps_preprocessor: gps data preprocessing pipeline
        plan_reprocessing_pipeline: plan preprocessing pipeline
        cluster_aggregator: gps aggregation pipeline
        spatial_joiner: performs spatial join of clusters & route plan
        spatial_validator: performs spatial validation of cluster & route plan join
        coverage_stats_extractor: performs extraction of coverage
            gps & plan against each other
        """
        self.gps_preprocessor = gps_preprocessor
        self.plan_preprocessor = plan_preprocessor
        self.cluster_aggregator = cluster_aggregator
        self.spatial_joiner = spatial_joiner
        self.spatial_validator = spatial_validator
        self.coverage_stats_extractor = coverage_stats_extractor

    def fit(self, X, y=None):
        return self

    def __preprocess_gps(self, data_container: DataContainer):
        data_container.gps = self.gps_preprocessor.transform(data_container.gps)
        return data_container

    def __preprocess_plan(self, data_container: DataContainer):
        data_container.plan = self.plan_preprocessor.transform(data_container.plan)
        return data_container

    def __compute_coverage(self, data_container: DataContainer):
        data_container.coverage_stats = self.coverage_stats_extractor.transform(data_container)
        return data_container

    def __aggregate_clusters(self, data_container: DataContainer):
        data_container.clusters = self.cluster_aggregator.transform(data_container.gps)
        return data_container

    def __join_clusters_and_plan(self, data_container: DataContainer):
        # TODO: get rid from SpatialJoinArguments in favor direct pass of data_container
        sjoin_args = SpatialJoinArguments(
            clustered_gps=data_container.clusters,
            route_plan=data_container.plan,
        )
        data_container.clusters_plan_join = self.spatial_joiner.transform(sjoin_args)
        return data_container

    def __validate_joins(self, data_container: DataContainer):
        data_container.clusters_plan_join = self.spatial_validator.transform(data_container.clusters_plan_join)
        return data_container

    def compute_coverage_stats(self, X: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        data_container = DataContainer.factory_instance(X)
        data_container = self.__preprocess_gps(data_container)
        data_container = self.__preprocess_plan(data_container)
        data_container = self.__compute_coverage(data_container)
        return data_container.coverage_stats

    def transform(self, X: Dict[str, pd.DataFrame]) -> pd.DataFrame:
        data_container = DataContainer.factory_instance(X)
        data_container = self.__preprocess_gps(data_container)
        data_container = self.__preprocess_plan(data_container)
        data_container = self.__compute_coverage(data_container)

        data_container = self.__aggregate_clusters(data_container)
        data_container = self.__join_clusters_and_plan(data_container)
        data_container = self.__validate_joins(data_container)
        return data_container



import pandas as pd
from pyproj import Transformer

from ..abstract import AbstractNode
from ..models import CRSProjectionModel
from ..models import DataFramePivotFields


class CRSTransformer(AbstractNode):
    """
    Module projecting coordinate reference system
    """

    def __init__(
        self,
        crs_projection: CRSProjectionModel,
        pivot_fields: DataFramePivotFields,
    ):
        self.crs_projection = crs_projection
        self.pivot_fields = pivot_fields
        self.crs_transformer = Transformer.from_crs(
            crs_projection.source_crs,
            crs_projection.target_crs,
            always_xy=True,
        )

    def fit(self, X, y=None):
        return self

    def __project_coordinates(self, df: pd.DataFrame) -> pd.DataFrame:
        return self.crs_transformer.transform(
            df[self.pivot_fields.source_lon],
            df[self.pivot_fields.source_lat],
        )

    def __assign_projections(self, df, projection):
        df[self.pivot_fields.projected_lon] = projection[0]
        df[self.pivot_fields.projected_lat] = projection[1]
        return df

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        projection = self.__project_coordinates(X)
        X = self.__assign_projections(X, projection)
        return X



from typing import Any
import geopandas as gpd
from pydantic import BaseModel

from ..models import DefaultValues


DEFAULT_VALUES = DefaultValues()


class SpatialJoinArguments(BaseModel):

    clustered_gps: Any
    route_plan: Any

    def __check_geometry(self, gdf: gpd.GeoDataFrame):
        if not isinstance(gdf, gpd.GeoDataFrame):
            raise ValueError("Input dataframe is not instance of gpd.GeoDataFrame")
        if not ("geometry" in gdf.columns):
            raise ValueError("geometry column is missing in one of the source dataframe")

    def validate_tables(self):
        self.__check_geometry(self.clustered_gps)
        self.__check_geometry(self.route_plan)



import pandas as pd


from ...abstract import AbstractNode
from ...models import DataContainer
from ...models import DataFramePivotFields
from ...models import DefaultValues


_PIVOT_FIELDS = DataFramePivotFields()
_DEFAULT_VALUES = DefaultValues()
_GROUP_COLUMNS = [
    _PIVOT_FIELDS.source_vehicle_id,
    _PIVOT_FIELDS.projected_date,
]
_AGG_COLUMN = "n_records"
_AGG_PARAMS = {
    _AGG_COLUMN: (_PIVOT_FIELDS.source_vehicle_id, "count"),
}
_GPS_SUFFIX = f"_{_DEFAULT_VALUES.sjoin_gps_suffix}"
_PLAN_SUFFIX = f"_{_DEFAULT_VALUES.sjoin_plan_suffix}"
_ACTION_FIELD = "action_required"

_DEFAULT_ACTION_MSG = "Keep as is"
_ACTION_MSG_TMP = "Drop vehicle-date in "


class CoverageStatistics(AbstractNode):
    """
    Table describes how provided gps covers provided route plan and inverse
    """

    def fit(self, X: DataContainer, y=None):
        return self

    def __agg_data(self, X: pd.DataFrame):
        return X.groupby(_GROUP_COLUMNS).agg(**_AGG_PARAMS).reset_index()

    def __add_action_recommendation(
        self,
        X: pd.DataFrame,
        source_suffix: str,
        target_suffix: str,
    ):
        drop_message = f"{_ACTION_MSG_TMP} {target_suffix}"
        is_missing_gps = X[_AGG_COLUMN + source_suffix].isna()
        X.loc[is_missing_gps, _ACTION_FIELD] = drop_message
        return X

    def __init_action_recommendation(
        self,
        X: pd.DataFrame,
    ):
        X[_ACTION_FIELD] = _DEFAULT_ACTION_MSG
        return X

    def transform(self, X: DataContainer):
        _gps_agg = self.__agg_data(X.gps)
        _plan_agg = self.__agg_data(X.plan)
        action_table = pd.merge(
            left=_gps_agg,
            right=_plan_agg,
            on=_GROUP_COLUMNS,
            how="outer",
            suffixes=[_GPS_SUFFIX, _PLAN_SUFFIX],
        )
        action_table = self.__init_action_recommendation(action_table)
        action_table = self.__add_action_recommendation(
            action_table,
            source_suffix=_GPS_SUFFIX,
            target_suffix=_PLAN_SUFFIX,
        )
        action_table = self.__add_action_recommendation(
            action_table,
            source_suffix=_PLAN_SUFFIX,
            target_suffix=_GPS_SUFFIX,
        )
        return action_table



from sklearn.pipeline import Pipeline

from .. import nodes as local_nodes
from ...abstract import AbstractPipelineFactory


class CoverageStatisticsFactory(AbstractPipelineFactory):
    @staticmethod
    def factory_pipeline() -> Pipeline:
        """
        Function creating instance of coverage statistics estimator
        instance pipeline of linker module

        Returns:
        sklear.pipeline.Pipeline: Scikit-learn pipeline
        """
        return Pipeline(
            [
                (
                    "coverage_stats_estimator",
                    local_nodes.CoverageStatistics(),
                ),
            ],
        )



from typing import Any, Dict
import pandas as pd
import pandera as pa
from pydantic import BaseModel


class CRSProjectionModel(BaseModel):
    source_crs: str = "EPSG:4326"
    target_crs: str = "EPSG:2326"


class DataFramePivotFields(BaseModel):

    source_lat: str = "lat"
    source_lon: str = "lon"
    source_datetime: str = "datetime"
    source_vehicle_id: str = "plate_no"

    projected_date: str = "date"
    projected_lat: str = "y"
    projected_lon: str = "x"
    computed_velocity: str = "computed_velocity"
    computed_unixtime: str = "unixtime"

    clustering_output: str = "cluster_id"
    fragmentation_output: str = "is_clustering_candidate"
    classification_output: str = "type_of_activity"

    sjoin_valid_flag: str = "sjoin_valid_flag"
    sjoin_spatial_dist: str = "sjoin_spatial_dist"
    sjoin_temporal_dist: str = "sjoin_temporal_dist"
    sjoin_overall_dist: str = "sjoin_overall_dist"

    gps_pk: str = "gps_primary_key"
    clusters_pk: str = "cluster_primary_key"
    plans_pk: str = "plans_primary_key"

    @property
    def pandera_schema(self):
        """
        Returns pandera schema
        """
        schema = {
            self.source_lat: pa.Column(float, coerce=True, nullable=False, required=True),
            self.source_lon: pa.Column(float, coerce=True, nullable=False, required=True),
            self.source_datetime: pa.Column(pd.Timestamp, coerce=True, nullable=False, required=True),
            self.source_vehicle_id: pa.Column(str, coerce=True, nullable=False, required=True),
            self.projected_date: pa.Column(pd.Timestamp, coerce=True, nullable=False, required=False),
            self.projected_lat: pa.Column(float, coerce=True, nullable=False, required=False),
            self.projected_lon: pa.Column(float, coerce=True, nullable=False, required=False),
            self.clustering_output: pa.Column(int, coerce=True, nullable=False, required=False),
            self.computed_unixtime: pa.Column(float, coerce=True, nullable=False, required=False),
            self.fragmentation_output: pa.Column(bool, coerce=True, nullable=False, required=False),
            self.classification_output: pa.Column(str, coerce=True, nullable=False, required=False),
            self.sjoin_valid_flag: pa.Column(bool, coerce=True, nullable=False, required=False),
            self.sjoin_spatial_dist: pa.Column(float, coerce=True, nullable=False, required=False),
            self.sjoin_temporal_dist: pa.Column(int, coerce=True, nullable=False, required=False),
            self.sjoin_overall_dist: pa.Column(float, coerce=True, nullable=False, required=False),
            self.gps_pk: pa.Column(str, coerce=True, nullable=False, required=False),
            self.clusters_pk: pa.Column(str, coerce=True, nullable=False, required=False),
            self.plans_pk: pa.Column(str, coerce=True, nullable=False, required=False),
        }
        schema = pa.DataFrameSchema(schema)
        return schema


class DefaultValues(BaseModel):
    noise_gps_cluster_id: int = -1
    sjoin_gps_suffix: str = "gps"
    sjoin_plan_suffix: str = "plan"
    sjoin_cluster_suffix: str = "cluster"
    pk_delimiter: str = "_"
    activity_linkage_gps_arg: str = "gps"
    activity_linkage_plan_arg: str = "plan"


class DataContainer(BaseModel):
    # NOTE: keys are needed to fabricate
    # instance with mandatory components
    gps_input_key: str = "gps"
    plan_input_key: str = "plan"
    gps: Any
    plan: Any

    coverage_stats: Any = None
    clusters: Any = None
    clusters_plan_join: Any = None

    @staticmethod
    def get_input(X: Dict[str, pd.DataFrame], key: str) -> pd.DataFrame:
        try:
            return X[key]
        except KeyError:
            message = f"Data are provided under incorrect {key} " "key to ActivityLinkageSession"
            raise KeyError(message)

    @classmethod
    def factory_instance(cls, X: Dict[str, pd.DataFrame]):

        keys = cls()
        gps = cls.get_input(X, keys.gps_input_key)
        plan = cls.get_input(X, keys.plan_input_key)

        return cls(gps=gps, plan=plan)

    def validated_coverage_stats(self):
        """
        Coverage stats validation is needed to ensure that all incoming clusters and plans
        are overallping over the days. it prevents bias of recall & precisions.
        """
        if self.coverage_stats.isna().any().any():
            msg = "You provided vehicle-date which is not listed in both source."
            msg += " Please, look at coverage stats, follow specified "
            msg += "action and recalculate linkage!"
            raise ValueError(msg)

    def __preprocess_clusters(self, pivots: DataFramePivotFields):
        _columns = [
            pivots.source_vehicle_id,
            pivots.projected_date,
            pivots.clustering_output,
            pivots.source_lon,
            pivots.source_lat,
            pivots.clusters_pk,
        ]
        return self.clusters.loc[:, _columns]

    def __preprocess_plan(self, pivots: DataFramePivotFields):
        _columns = [
            pivots.source_vehicle_id,
            pivots.projected_date,
            pivots.source_lon,
            pivots.source_lat,
            pivots.plans_pk,
        ]
        plan = self.plan[_columns]
        plan = plan.rename(
            columns={
                pivots.source_lat: f"{pivots.source_lat}_plan",
                pivots.source_lon: f"{pivots.source_lon}_plan",
            },
        )
        return plan

    def get_concatenated_gps(self) -> pd.DataFrame:
        """
        Function left joins (route plan details, clusters) to gps records

        Returns:
        pd.DataFrame: concatenated gps
        """
        _pivots = DataFramePivotFields()
        _defaults = DefaultValues()
        clusters = self.__preprocess_clusters(_pivots)
        plan = self.__preprocess_plan(_pivots)
        return (
            self.gps.merge(
                clusters,
                how="left",
                on=[
                    _pivots.source_vehicle_id,
                    _pivots.projected_date,
                    _pivots.clustering_output,
                ],
                suffixes=[
                    f"_{_defaults.sjoin_gps_suffix}",
                    f"_{_defaults.sjoin_cluster_suffix}",
                ],
            )
            .merge(self.clusters_plan_join, how="left", on=[_pivots.clusters_pk])
            .merge(
                plan,
                how="left",
                on=[_pivots.plans_pk],
                suffixes=["", f"_{_defaults.sjoin_plan_suffix}"],
            )
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gps_activity',
 'gps_activity.extraction',
 'gps_activity.extraction.factory',
 'gps_activity.extraction.factory.classifiers',
 'gps_activity.extraction.factory.clustering',
 'gps_activity.extraction.factory.fragmentation',
 'gps_activity.extraction.nodes',
 'gps_activity.linker',
 'gps_activity.linker.factory',
 'gps_activity.linker.nodes',
 'gps_activity.metrics',
 'gps_activity.metrics.nodes',
 'gps_activity.nodes']

package_data = \
{'': ['*']}

install_requires = \
['Rtree>=1.0.0,<2.0.0',
 'geopandas>=0.11.1,<0.12.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'pandera>=0.11.0,<0.12.0',
 'pygeos>=0.12.0,<0.13.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'gps-activity',
    'version': '0.7.0',
    'description': 'A light-weight mobile module for analysis of GPS activity',
    'long_description': '# **GPS activity** 🚛\n\n![Python versions](https://img.shields.io/pypi/pyversions/gps_activity)\n![Latest release](https://img.shields.io/github/v/release/WasteLabs/gps_activity)\n![Latest release date](https://img.shields.io/github/release-date/WasteLabs/gps_activity)\n![License](https://img.shields.io/github/license/WasteLabs/gps_activity)\n[![CI test](https://github.com/WasteLabs/gps_activity/actions/workflows/ci-tests.yaml/badge.svg)](https://github.com/WasteLabs/gps_activity/actions/workflows/ci-tests.yaml)\n[![codecov](https://codecov.io/gh/WasteLabs/gps_activity/branch/main/graph/badge.svg?token=58AY7B1YBB)](https://codecov.io/gh/WasteLabs/gps_activity)\n\nA light-weight module for analysis of GPS activity. Package is designed to be trade-off solution for both researchers and developers in Waste Labs. Using `gps_activity` you can:\n\n1. Cluster your time-series gps records to extract activity points\n2. Join activity points with original plan or operation report\n3. Estimate clustering performance\n\n## Installation ☁️ -> 🖥️\n\nUsing pip:\n\n```bash\npip3 install gps_activity\n```\n\n## **Python package modules** 📦\n\n---\n\n- **extraction**: clusters GPS records and extracts cluster activities ([checkout module structure]((https://github.com/WasteLabs/gps_activity/tree/main/docs/extraction/README.md)))\n- **linker**: joins route plan and clustered gps records\n- **metrics**: estimates clustering performance based on:\n    1. internal source: ones that based on inter & intra cluster distances\n    2. external source: joined route plan and clustered gps records (output of **linker** module)\n\n![components](docs/diagrams/gps_activity_components.png)\n\n## **Extraction modules** 🔵 🟣 ⚫️\n\n---\n\nOrganized by preprocessing, fragmentation & clustering steps that ultimately are packed into `ActivityExtractionSession` object and orchestrated.\n\n### ⚠️ `ActivityExtractionSession` assumptions and constrains\n\n1. 1 vehicle = 1 session run: to avoid clusters overlap\n1. GPS records must be ordered over datetime: ensures adjacent consistency of gps\n1. No duplicated gps records over vehicle-timstamp: avoids division by zero during computing `velocity`\n\n### 🚀 **VHFDBSCAN**: Velocity hardlimit fragmentation Density-based spatial clustering of applications with noise\n\n- Fragmentation is performing by applying hardlimiting on velocity computed from `lat`, `lon` and `datetime` columns\n- Clustering is performed by classical DBSCAN that considers non-cluster candidates as noise\n\n```python\nfrom gps_activity import ActivityExtractionSession\nfrom gps_activity.extraction.factory.preprocessing import PreprocessingFactory\nfrom gps_activity.extraction.factory.fragmentation import VelocityFragmentationFactory\nfrom gps_activity.extraction.factory.clustering import FDBSCANFactory\n\n\npreprocessing = PreprocessingFactory.factory_pipeline(\n    source_lat_column="lat",\n    source_lon_column="lon",\n    source_datetime="datetime",\n    source_vehicle_id="plate_no",\n    source_crs="EPSG:4326",\n    target_crs="EPSG:2326",\n)\n\nfragmentation = VelocityFragmentationFactory.factory_pipeline(max_velocity_hard_limit=4)\nclustering = FDBSCANFactory.factory_pipeline(eps=30, min_samples=3)\n\nactivity_extraction = ActivityExtractionSession(\n    preprocessing=preprocessing,\n    fragmentation=fragmentation,\n    clustering=clustering,\n)\n\nactivity_extraction.predict(gps)\n```\n\n### 🚀 **VHFSTCM**: Velocity hardlimit fragmentation spatio-temporal clustering model\n\n- Fragmentation is performing by applying hardlimiting on velocity computed from `lat`, `lon` and `datetime` columns\n- Clustering is performed according steps:\n    1. Generated adjacent proximity mask (if cluster pair distance <= `eps`)\n    2. Clusters ID are generated according: proximity mask and fragmentation flag\n    3. GPS records grouped by `cluster_id` and aggregated cluster time span\n    4. Cluster is validated if time span >= `min_duration_sec`\n    5. Validated cluster ids are set to original GPS records\n\n```python\nfrom gps_activity import ActivityExtractionSession\nfrom gps_activity.extraction.factory.preprocessing import PreprocessingFactory\nfrom gps_activity.extraction.factory.fragmentation import VelocityFragmentationFactory\nfrom gps_activity.extraction.factory.clustering import STCMFactory\n\n\npreprocessing = PreprocessingFactory.factory_pipeline(\n    source_lat_column="lat",\n    source_lon_column="lon",\n    source_datetime="datetime",\n    source_vehicle_id="plate_no",\n    source_crs="EPSG:4326",\n    target_crs="EPSG:2326",\n)\n\nfragmentation = VelocityFragmentationFactory.factory_pipeline(max_velocity_hard_limit=4)\nclustering = STCMFactory.factory_pipeline(\n    source_vehicle_id_column="plate_no",\n    eps=30,\n    min_duration_sec=60\n)\n\nstcm = ActivityExtractionSession(\n    preprocessing=preprocessing,\n    fragmentation=fragmentation,\n    clustering=clustering,\n)\n```\n\n## Linker module implementation 🔵 🟣 ⚫️\n\n**[Overview linker module components](https://github.com/WasteLabs/gps_activity/tree/main/docs/linker/README.md)**\n\n```python\n# Initilize linkage components\nfrom gps_activity import ActivityLinkageSession\nfrom gps_activity.linker.factory import PreprocessingFactory\nfrom gps_activity.linker.factory import ClusterAggregationFactory\nfrom gps_activity.linker.factory import JoinValidatorFactory\nfrom gps_activity.linker.factory import SpatialJoinerFactory\nfrom gps_activity.linker.factory import CoverageStatisticsFactory\n\n\nMAX_DISTANCE = 100\nMAX_DAYS_DISTANCE = 1\n\n\ngps_link_preprocess_pipeline = PreprocessingFactory.factory_pipeline(\n    source_lat_column="lat",\n    source_lon_column="lon",\n    source_datetime="datetime",\n    source_vehicle_id="plate_no",\n    source_crs=WSG_84,\n    target_crs=HK_CRS,\n    generate_primary_key_for="gps",\n    source_composite_keys=["plate_no", "datetime", "lat", "lon"],\n)\n\nplans_link_preprocess_pipeline = PreprocessingFactory.factory_pipeline(\n    source_lat_column="lat",\n    source_lon_column="lng",\n    source_datetime="datetime",\n    source_vehicle_id="re-assigned by Ricky",\n    source_crs=WSG_84,\n    target_crs=HK_CRS,\n    generate_primary_key_for="plan",\n    source_composite_keys=["CRN#"],\n)\n\ncluster_agg_pipeline = ClusterAggregationFactory.factory_pipeline(\n    source_lat_column="lat",\n    source_lon_column="lon",\n    source_datetime="datetime",\n    source_vehicle_id="plate_no",\n    source_crs=WSG_84,\n    target_crs=HK_CRS,\n)\n\nspatial_joiner = SpatialJoinerFactory.factory_pipeline(how="inner", max_distance=MAX_DISTANCE)\nspatial_validator = JoinValidatorFactory.factory_pipeline(max_days_distance=MAX_DAYS_DISTANCE,\n                                                          ensure_vehicle_overlap=True)\ncoverage_stats_extractor = CoverageStatisticsFactory.factory_pipeline()\n\n\ngps_linker_session = ActivityLinkageSession(\n    gps_preprocessor=gps_link_preprocess_pipeline,\n    plan_preprocessor=plans_link_preprocess_pipeline,\n    cluster_aggregator=cluster_agg_pipeline,\n    spatial_joiner=spatial_joiner,\n    spatial_validator=spatial_validator,\n    coverage_stats_extractor=coverage_stats_extractor,\n)\n\n\nlinker_results = gps_linker_session.transform({\n    "gps": clustered_gps,\n    "plan": plans,\n})\n```\n\n## Metrics module implementation 📊\n\n* **NOTE**: This module is highly experimental\n* **NOTE**: This module depends on `linker` module\n\n```python\nfrom gps_activity.metrics import ActivityMetricsSession\nfrom gps_activity.metrics.models import Metrics\n\n\nmetrics = ActivityMetricsSession(beta=2)\nmetrics = metrics.transform(linker_results)\n```\n',
    'author': 'Adil Rashitov',
    'author_email': 'adil@wastelab.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from ...abstract import AbstractPredictor
from ...models import DataFramePivotFields
from ...models import DefaultValues


pivot_fields = DataFramePivotFields()
defaults = DefaultValues()


class FDBSCAN(AbstractPredictor):
    """
    Fragment Density Based Spatial Clustering, where
    fragmentation is selection of potential clustering candidates
    """

    SKIP_CANDIDATES_DEFAULT = defaults.noise_gps_cluster_id
    TEMP_CLUSTER_COL = pivot_fields.clustering_output

    # flake8: noqa: CFQ002
    def __init__(
        self,
        clustering_candidate_col: str,
        eps: float = 0.5,
        min_samples: float = 5,
        metric: str = "euclidean",
        metric_params=None,
        algorithm="auto",
        leaf_size=30,
        p=None,
        n_jobs=-1,
    ):
        self.clustering_candidate_col = clustering_candidate_col
        self.eps = eps
        self.min_samples = min_samples
        self.metric = metric
        self.metric_params = metric_params
        self.algorithm = algorithm
        self.leaf_size = leaf_size
        self.p = p
        self.n_jobs = n_jobs
        self.__dbscan = DBSCAN(
            eps=eps,
            min_samples=min_samples,
            metric=metric,
            metric_params=metric_params,
            algorithm=algorithm,
            leaf_size=leaf_size,
            p=leaf_size,
            n_jobs=n_jobs,
        )

    def __get_model_inputs(self, X: pd.DataFrame):
        return X.loc[self.__mask, self.__columns]

    def __extract_clustering_mask(self, X: pd.DataFrame):
        self.__mask = X[self.clustering_candidate_col].copy()

    def __extract_clustering_columns(self, X: pd.DataFrame):
        X = X.drop(columns=[self.clustering_candidate_col]).copy()
        self.__columns = list(X.columns)

    def fit(self, X: pd.DataFrame, y=None):
        self.__extract_clustering_mask(X)
        self.__extract_clustering_columns(X)

    def __impute_noise(self, X: pd.DataFrame) -> pd.DataFrame:
        column = self.TEMP_CLUSTER_COL
        X.loc[:, column] = X.loc[:, column].fillna(self.SKIP_CANDIDATES_DEFAULT)
        return X

    def __assign_predictions(self, X: pd.DataFrame, y: np.ndarray) -> pd.DataFrame:
        X.loc[self.__mask, self.TEMP_CLUSTER_COL] = y
        return X

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        X = self._get_input_copy(X)
        self.fit(X)
        y = self.__dbscan.fit_predict(self.__get_model_inputs(X))
        X = self.__assign_predictions(X=X, y=y)
        X = self.__impute_noise(X)
        return X[self.TEMP_CLUSTER_COL].values

    def fit_predict(self, X: pd.DataFrame, y=None):
        return self.predict(X)



import pandas as pd

from ..abstract import AbstractNode


class UnixtimeExtractor(AbstractNode):
    """
    Module adding unixtime to dataframe
    """

    SECONDS_DENOMINATOR = 10**9

    def __init__(
        self,
        source_column: str,
        target_column: str,
    ):
        self.source_column = source_column
        self.target_column = target_column

    def fit(self, X, y=None):
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        X[self.source_column] = pd.to_datetime(X[self.source_column])
        X[self.target_column] = X[self.source_column].astype(int) / self.SECONDS_DENOMINATOR
        return X



from ...abstract import AbstractNode
from ...models import DataContainer
from ...models import DataFramePivotFields
from .precision import Precision
from .recall import Recall

_PIVOT_FIELDS = DataFramePivotFields()
_PRIMARY_KEY = _PIVOT_FIELDS.clusters_pk


class Fbeta(AbstractNode):
    def __init__(self, beta: float = 1):
        self.precision = Precision()
        self.recall = Recall()
        self.beta = beta

    def fit(self, X: DataContainer, y=None):
        return self

    def __fbeta(self, recall: float, precision: float, beta: float) -> float:
        _numerator = (1 + beta**2) * precision * recall
        _denominator = beta**2 * precision + recall
        return _numerator / _denominator

    def transform(self, X: DataContainer):
        _precision = self.precision.transform(X)
        _recall = self.recall.transform(X)
        fbeta_score = self.__fbeta(
            recall=_recall,
            precision=_precision,
            beta=self.beta,
        )
        return fbeta_score



from .extraction import ActivityExtractionSession
from .linker import ActivityLinkageSession
from .metrics import ActivityMetricsSession


__all__ = [
    "ActivityExtractionSession",
    "ActivityLinkageSession",
    "ActivityMetricsSession",
]



from typing import Union

from geopandas import gpd
import pandas as pd

from ...abstract import AbstractNode
from ...linker.models import SpatialJoinArguments
from ...models import DataFramePivotFields
from ...models import DefaultValues


defaults = DefaultValues()
pivot_fields = DataFramePivotFields()
_GPS_PREFIX = defaults.sjoin_gps_suffix
_PLAN_PREFIX = defaults.sjoin_plan_suffix
_L2_DISTANCE_COLUMN = pivot_fields.sjoin_spatial_dist


class SpatialJoiner(AbstractNode):
    def __init__(
        self,
        how: str = "inner",
        max_distance: Union[float, int] = 80,
    ):
        self.how = how
        self.max_distance = max_distance

    def fit(self, X: pd.DataFrame, y=None):
        return self

    def transform(self, X: SpatialJoinArguments):
        X.validate_tables()

        sjoin = gpd.sjoin_nearest(
            left_df=X.clustered_gps,
            right_df=X.route_plan,
            how=self.how,
            max_distance=self.max_distance,
            lsuffix=_GPS_PREFIX,
            rsuffix=_PLAN_PREFIX,
            distance_col=_L2_DISTANCE_COLUMN,
        )

        return sjoin
